import json

from aiospb.messages import Payload, PayloadEncoder
from aiospb.shared import Metric


class JsonPayloadEncoder(PayloadEncoder):
    """Encoder of messages for testing purposes"""

    def encode(self, payload: Payload) -> bytes:
        """Convert a payload dict to a string for publishing"""
        if payload.online is not None:
            data = {"timestamp": payload.timestamp, "online": payload.online}
        else:
            data = {"timestamp": payload.timestamp, "seq": payload.seq}
            data["metrics"] = [metric.as_dict() for metric in payload.metrics]

        return json.dumps(data).encode("utf-8")

    def decode(self, b: bytes) -> Payload:
        """Convert payload to a payload dict"""

        data = json.loads(b.decode("utf-8"))
        if "online" in data:
            return Payload(data["timestamp"], online=data["online"])

        metrics = [Metric.from_dict(item) for item in data["metrics"]]
        return Payload(timestamp=data["timestamp"], metrics=metrics, seq=data["seq"])



__all__ = (
    "DeviceConnectionIsBroken",
    "EdgeNode",
    "Historian",
    "MetricInfo",
    "MetricNotFound",
    "MetricsNetwork",
    "NodeDevice",
    "Reading",
    "ReadingStore",
)

from .devices import DeviceConnectionIsBroken, MetricsNetwork, NodeDevice
from .nodes import EdgeNode
from .ports import Historian, MetricInfo, MetricNotFound, Reading, ReadingStore



"""Implementation of ports with file system (files and directories)"""
import os
import os.path

import aiofiles

from aiospb.nodes.ports import Historian

from ..shared import DataType, DeviceKey, Metric, Quality


class FSHistorian(Historian):
    """Store historical changes in plain files from a operating system"""

    def __init__(self, directory: str):
        self._dir = directory

    def _get_fn(self, key: DeviceKey) -> str:
        return os.path.join(self._dir, key.path.replace("/", "--") + ".csv")

    def _create_line(self, metric: Metric) -> str:
        quality = (
            metric.properties["quality"].value if "quality" in metric.properties else ""
        )
        return f"{metric.timestamp};{metric.value};{metric.data_type.name};{quality};{metric.alias};{metric.name}\n"

    async def load(self, key: DeviceKey) -> list[Metric]:
        metrics = []
        try:
            async with aiofiles.open(self._get_fn(key), "r") as f:
                while True:
                    line = await f.readline()

                    values = line[:-1].split(";")

                    if len(values) == 1:
                        break

                    ts, value, datatype, quality, alias, name = values
                    m = Metric(
                        int(ts),
                        DataType[datatype].convert_value(value),
                        DataType[datatype],
                        int(alias),
                        name,
                        is_historical=True,
                    )
                    if quality:
                        m.quality = Quality(int(quality))
                    metrics.append(m)
        except FileNotFoundError:
            ...

        return metrics

    async def save(self, key: DeviceKey, metrics: list[Metric]):
        fn = os.path.join(self._dir, key.path.replace("/", "--") + ".csv")
        async with aiofiles.open(fn, "a") as f:
            lines = [self._create_line(metric) for metric in metrics]
            await f.writelines(lines)

    async def clear(self, key: DeviceKey):
        os.remove(self._get_fn(key))

    def get_device_keys(self, node_key: DeviceKey) -> list[DeviceKey]:
        keys = []
        for fn in os.listdir(self._dir):
            if "." in fn:
                key = fn[: fn.index(".")]
                key = key.replace("--", "/")
                device_key = DeviceKey(key)
                if device_key.node == node_key:
                    keys.append(device_key)
        return keys



import abc
import base64
import tempfile
from dataclasses import dataclass
from enum import Enum
from typing import Any, Self

from ..data import Metric


def matches_topic(pattern: str, value: str) -> bool:
    """
    Recursively checks if a topic matches a pattern with + and # wildcards.

    :param pattern: The pattern to match against (e.g., "home/+/temperature").
    :param value: The topic to check (e.g., "home/livingroom/temperature").
    :return: True if the topic matches the pattern, False otherwise.
    """
    # Split the topic and pattern into parts based on '/'
    topic_parts = value.split("/")
    pattern_parts = pattern.split("/")

    # Base case: if both topic and pattern are empty, they match
    if not topic_parts and not pattern_parts:
        return True

    # If one is empty and the other is not, they don't match
    if not topic_parts or not pattern_parts:
        return False

    # Get the current level of the topic and pattern
    current_topic = topic_parts[0]
    current_pattern = pattern_parts[0]

    # If the current pattern is '#', it matches the rest of the topic
    if current_pattern == "#":
        return True

    # If the current topic match
    if current_pattern == "+" or current_pattern == current_topic:
        return matches_topic("/".join(pattern_parts[1:]), "/".join(topic_parts[1:]))

    try:
        if re.fullmatch(current_pattern, current_topic):
            return matches_topic("/".join(pattern_parts[1:]), "/".join(topic_parts[1:]))
    except re.error:
        # If the pattern is not a valid regex, treat it as a literal string
        if current_pattern == current_topic:
            return matches_topic("/".join(pattern_parts[1:]), "/".join(topic_parts[1:]))

    # If none of the above, they don't match
    return False


@dataclass
class Topic:
    value: str

    @property
    def component_name(self) -> str:
        items = self.value.split("/")
        if items[1] == "STATE":
            return "/".join(items[2:])
        else:
            return "/".join([items[1]] + items[3:])

    @property
    def message_type(self) -> str:
        items = self.value.split("/")
        if items[1] == "STATE":
            return "STATE"
        return items[2]

    @classmethod
    def from_component(cls, name: str, message_type: str) -> Self:
        if message_type not in (
            "STATE",
            "NDATA",
            "NBIRTH",
            "NDEATH",
            "NCMD",
            "DDATA",
            "DBIRTH",
            "DDEATH",
        ):
            raise ValueError(f'Message type "{message_type}" is not a standard')
        if message_type == "STATE":
            return cls(f"spBv1.0/STATE/{name}")
        items = name.split("/")
        return cls(f"spBv1.0/{items[0]}/{message_type}/{'/'.join(items[1:])}")


@dataclass
class Payload(abc.ABC):
    timestamp: int

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""

    @abc.abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""


class MessageType(Enum):
    STATE = "STATE"
    NBIRTH = "NBIRTH"
    NDATA = "NDATA"
    NDEATH = "NDEATH"
    NCMD = "NCMD"
    DBIRTH = "DBIRTH"
    DDATA = "DDATA"
    DDEATH = "DDEATH"
    DCMD = "DCMD"


class ComponentType:
    HOST = "HOST"
    SP_DEVICE = "SP_DEVICE"
    EDGE_NODE = "EDGE_NODE"


@dataclass
class SpbMessage:
    component_name: str
    message_type: MessageType
    payload: Payload

    def __pos_init__(self):
        ...

    @property
    def topic(self) -> str:
        """Message type"""
        items = self.component_name.split("/")
        if len(items) == 1:
            return f"spBv1.0/STATE/{items[0]}"
        elif len(items) == 2:
            return f"spBv1.0/{items[0]}/{self.message_type}/{items[1]}"
        elif len(items) == 3:
            return f"spBv1.0/{items[0]}/{self.message_type}/{items[1]}/{items[3]}"

        raise ValueError(f"Component name {self.component_name} has level over 3")

    @property
    def component_type(self) -> ComponentType:
        return {
            0: ComponentType.HOST,
            1: ComponentType.EDGE_NODE,
            2: ComponentType.SP_DEVICE,
        }[self.component_name.count("/")]


@dataclass
class Will:
    """Will message to sent to MQTT broker"""

    message: SpbMessage
    qos: int
    retain: bool


class MqttClient(abc.ABC):
    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Is the client connected to the Broker?"""

    @abc.abstractmethod
    async def connect(self, component_name: str, will: Will):
        """Connect a component to MQTT server"""

    @abc.abstractmethod
    async def publish(self, message: SpbMessage, qos: int, retain: bool):
        """Publish a message  to the topic"""

    @abc.abstractmethod
    async def deliver_message(self) -> SpbMessage:
        """Return a messsage recieved from the MQTT Server"""

    @abc.abstractmethod
    async def subscribe(self, topic: str, qos: int):
        """Subscribe the component to recieve messages from a topic"""

    @abc.abstractmethod
    async def disconnect(self):
        """Disconnect the client from the MQTT server"""


@dataclass
class HostPayload(Payload):
    """Payload to send by MQTT for State messages"""

    online: bool

    @classmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""
        return cls(payload_map["timestamp"], payload_map["online"])

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""
        return {"timestamp": self.timestamp, "online": self.online}


@dataclass
class MessageContent(abc.ABC):
    """Interface of a message content"""

    @abc.abstractmethod
    async def send(
        self,
        mqtt_client: "MqttClient",
        timestamp: int,
        component_name: str,
        seq: int | None = None,
    ):
        """Publish to MQTT Broker the content of the message"""


@dataclass
class NodePayload(Payload):
    """Payload to send by MQTT for/to edge node messages"""

    metrics: list[Metric]
    seq: int | None = None

    @classmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""
        return cls(
            payload_map["timestamp"],
            [Metric.from_dict(value) for value in payload_map["metrics"]],
            payload_map.get("seq"),
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""
        outcome = {
            "timestamp": self.timestamp,
            "metrics": [dto.as_dict() for dto in self.metrics],
        }
        if self.seq is not None:
            outcome["seq"] = self.seq

        return outcome


@dataclass
class MqttConfig:
    hostname: str
    port: int
    credentials: str = ""
    ca_cert: str = ""
    keepalive: int = 30

    def __post_init__(self):
        self.port = int(self.port)

    def deploy_certificate_file(self) -> str:
        """Save a certificate content to a file and return its filename"""
        if not self.ca_cert:
            raise ValueError("There is no certificate to save")

        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            content = "-----BEGIN CERTIFICATE-----\n"
            for line in range(500):
                if (line + 1) * 64 > len(self.ca_cert):
                    content += self.ca_cert[line * 64 :] + "\n"
                    break

                content += self.ca_cert[line * 64 : (line + 1) * 64] + "\n"

            content += "-----END CERTIFICATE-----"
            f.write(content)
            return f.name

    def login_info(self) -> tuple[str | None, str | None]:
        """Return username and password for login to the MQTT Server"""

        if not self.credentials:
            return (None, None)

        tokens = (base64.b64decode(self.credentials).decode()).split(":")
        return (tokens[0], ":".join(tokens[1:]))


class MessageEncoder(abc.ABC):
    """Encode the message before sending it as payload in the message"""

    @abc.abstractmethod
    def encode(self, payload: Payload) -> bytes:
        """Convert a message to a payload"""

    @abc.abstractmethod
    def decode(self, payload: bytes) -> Payload:
        """Convert payload to a message object"""


class MqttError(Exception):
    """Wraps error of comunications by any MQTT adapter"""



import asyncio
import logging
from collections.abc import Callable, Coroutine
from copy import deepcopy

from aiospb.messages import Event, Message
from aiospb.shared import (
    Clock,
    DataType,
    DeviceKey,
    Metric,
    Quality,
    SparkplugBError,
    UtcClock,
)

from .scanning import Scanner

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from .ports import DeviceConnectionIsBroken, MetricsNetwork, ReadingStore
from .values import MetricInfo, Reading


class NodeDevice:
    """Node Device publish device messages and execute commmands"""

    def __init__(
        self,
        key: DeviceKey,
        metrics_net: MetricsNetwork,
        reading_store: ReadingStore,
        clock: Clock | None = None,
    ):
        self._key = key
        self._control_group = (
            "Device Control/" if self._key.device_name else "Node Control/"
        )
        self._net = metrics_net
        self._readings = reading_store
        self._notify_message = None
        self._clock = clock or UtcClock()

        self._scanners = {}
        self._scan_rate = None
        self._born = False

    @property
    def key(self) -> DeviceKey:
        return self._key

    def add_node_notification(
        self,
        notify_message: Callable[[Message], Coroutine[None, None, None]],
    ):
        self._notify_message = notify_message

    async def rise(self, default_scan_rate: int):
        if not self._notify_message:
            raise ValueError(
                f"Device {self._key} can not start without node to notify messages"
            )

        self._scan_rate = default_scan_rate
        for scanner in self._scanners.values():
            scanner.stop()
        self._scanners.clear()

        dev_infos = await self._net.get_info(self._key)
        if not dev_infos:
            raise Exception(f"Device {self._key} has not metrics to report")

        await self._net.connect_device(self.key)
        metrics = await self._read_birth_metrics(dev_infos, default_scan_rate * 2)
        ts = self._clock.timestamp()
        controls = [
            Metric(ts, False, DataType.Boolean, name=self._control_group + "Rebirth"),
            Metric(ts, False, DataType.Boolean, name=self._control_group + "Restart"),
            Metric(
                ts,
                default_scan_rate,
                DataType.Int32,
                name=self._control_group + "Scan Rate",
            ),
        ]
        if not self._key.device_name:
            controls.append(
                Metric(ts, False, DataType.Boolean, name="Node Control/Reboot")
            )

        for index in range(
            len(metrics)
        ):  # Replace Scan Rate if it is reported by the net
            if metrics[index].name == self._control_group + "Scan Rate":
                self._scan_rate = metrics[index].value
                controls[2] = metrics[index]
                del metrics[index]
                break

        metrics = controls + metrics
        scan_plan = self._construct_scan_plan(metrics)
        for scan_rate, aliases in scan_plan.items():
            if aliases:
                self._scanners[scan_rate] = scanner = Scanner(
                    self._key,
                    aliases,
                    self._net,
                    self._readings,
                    self._recieve_metric_changes,
                    self._clock,
                )
                scanner.start(scan_rate)

        ts = self._clock.timestamp()
        event = Event.DEVICE_IS_BORN if self._key.device_name else Event.NODE_IS_BORN
        await self._notify_message(Message(event, self._key, metrics))
        self._born = True

    def _construct_scan_plan(self, metrics: list[Metric]) -> dict[int, list[int]]:
        scan_rate = metrics[2].value
        if type(scan_rate) is not int:
            raise ValueError(f"Scan Rate shall be integer, not {type(scan_rate)}")

        plan = {scan_rate: []}
        for m in metrics:
            if not m.alias:  # Decided scan only if exist alias
                continue
            if m.properties and "scan_rate" in m.properties:  # Special metrics
                sr = m.properties["scan_rate"]
                if sr.value == 0:
                    continue

                if sr.value not in plan:
                    plan[sr.value] = []

                plan[sr.value].append(m.alias)

            plan[scan_rate].append(m.alias)

        return plan

    async def _read_birth_metrics(
        self, infos: list[MetricInfo], timeout_ms: int
    ) -> list[Metric]:
        tasks = [
            asyncio.create_task(self._read_birth_metric(i, timeout_ms)) for i in infos
        ]
        await asyncio.wait(tasks)
        metrics = []
        readings = {}
        for task in tasks:
            metric, reading = task.result()
            if metric.alias:
                readings[metric.alias] = reading
            metrics.append(metric)

        await self._readings.set(readings)
        return metrics

    async def _read_birth_metric(
        self, i: MetricInfo, timeout_ms: int
    ) -> tuple[Metric, Reading]:
        try:
            value = await self._clock.wait_for(
                asyncio.create_task(
                    self._net.read(self._key, alias=i.alias, name=i.name), name=i.name
                ),
                timeout_ms,
            )

            reading = Reading(value, i.data_type, Quality.GOOD)
        except TimeoutError:
            reading = Reading(None, i.data_type, Quality.STALE)
        except DeviceConnectionIsBroken:
            reading = Reading(None, i.data_type, Quality.STALE)
        except Exception as e:
            logger.error(f'Reading birth metric of "{i.name}"')
            logger.exception(e)
            reading = Reading(None, i.data_type, Quality.BAD)

        return (i.create_birth_metric(self._clock.timestamp(), reading), reading)

    async def _recieve_metric_changes(self, metrics: list[Metric]):
        if not self._notify_message:
            raise self._no_observer_exception()

        if not self._born:
            raise SparkplugBError(
                f"Stoping sent of changes, device {self._key} hasn't been born"
            )
        event = (
            Event.DEVICE_DATA_HAS_CHANGED
            if self._key.device_name
            else Event.NODE_DATA_HAS_CHANGED
        )
        await self._notify_message(Message(event, self._key, metrics))

    async def die(self):
        for scanner in self._scanners.values():
            scanner.stop()
        self._scanners.clear()

        if not self._notify_message:
            raise self._no_observer_exception()

        event = Event.DEVICE_HAS_DIED if self._key.device_name else Event.NODE_HAS_DIED
        await self._notify_message(Message(event, self._key))
        await self._net.disconnect_device(self.key)
        self._born = False

    def _no_observer_exception(self):
        return SparkplugBError("There is no EdgeNode observing device")

    async def write_metrics(self, metrics: list[Metric], timeout_ms: int):
        """Write to device the metrics in order"""
        if not self._born:
            raise SparkplugBError(f"Device {self._key} has't been born")

        names = [m.name for m in metrics]
        if f"{self._control_group}Scan Rate" in names:
            val = metrics[names.index(f"{self._control_group}Scan Rate")].value
            if type(val) is int:
                await self._set_scan_rate(val)

        idx = 0
        readings = {}
        metrics = deepcopy(metrics)
        dead_line = self._clock.timestamp() + timeout_ms
        for idx, metric in enumerate(metrics):
            try:
                task = asyncio.create_task(
                    self._net.write(
                        self._key, metric.value, alias=metric.alias, name=metric.name
                    )
                )
                timeout_ms = dead_line - self._clock.timestamp()
                if timeout_ms <= 0:
                    raise TimeoutError()
                alias = await self._clock.wait_for(task, timeout_ms)

                if alias:
                    readings[alias] = Reading(
                        metric.value, metric.data_type, quality=Quality.GOOD
                    )
                metric.timestamp = self._clock.timestamp()
                metric.quality = Quality.GOOD
            except TimeoutError:
                metric.timestamp = self._clock.timestamp()
                metric.quality = Quality.STALE
            except Exception as e:
                logger.error(
                    f"Writing to {metric.name}:{metric.alias}, value {metric.value}"
                )
                logger.exception(e)
                metric.timestamp = self._clock.timestamp()
                metric.quality = Quality.BAD
                break

        if idx < len(metrics) - 1:
            for metric in metrics[idx:]:
                metric.timestamp = self._clock.timestamp()
                metric.quality = Quality.BAD

        await self._recieve_metric_changes(metrics)
        await self._readings.set(readings)

    async def _set_scan_rate(self, value: int):
        if value == self._scan_rate:
            return
        scanner = self._scanners.pop(self._scan_rate)
        if scanner:
            scanner.stop()
            scanner.start(value)
